import com.urbancode.anthill3.AHPTool
import java.util.regex.Matcher
import org.xml.sax.InputSource
import javax.xml.parsers.SAXParserFactory
import org.xml.sax.helpers.DefaultHandler
import java.util.regex.Pattern
import org.xml.sax.Attributes
import org.xml.sax.SAXException
import groovy.xml.StreamingMarkupBuilder
import com.urbancode.anthill3.CommandHelper

class FileSet {
    def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
    def base
    def includes = []
    def excludes = []

    def FileSet(base) {
        if (base instanceof File) {
            this.base = base
        }
        else if (base == null) {
            this.base = new File('.').absoluteFile
        }
        else {
            this.base = new File(base)
        }
    }

    def include(antPattern) {
        forNonEmptyLines(antPattern) {
            includes << convertToPattern(it)
        }
    }

    def exclude(antPattern) {
        forNonEmptyLines(antPattern) {
            excludes << convertToPattern(it)
        }
    }

    def each(Closure closure) {
        base.eachFileRecurse { file ->
            def path = file.path.replace('\\', '/').substring(base.path.length())
            def matches = false
            for (p in includes) {
                if (path =~ p) {
                    matches = true
                    break;
                }
            }
            if (matches) {
                for (p in excludes) {
                    if (path =~ p) {
                        matches = false;
                        break;
                    }
                }
            }
            if (matches) {
                closure(file)
            }
        }
    }

    def files() {
        def list = []
        each { list << it }
        return list
    }

    private forNonEmptyLines(strings, Closure closure) {
        if (strings instanceof Collection) {
            strings.each {
                def trimmed = it?.trim()
                if (trimmed?.length() > 0) {
                    closure(it)
                }
            }
        }
        else if (strings != null) {
            forNonEmptyLines(strings.readLines(), closure)
        }
    }

    private convertToPattern(antPattern) {
        // trim leading / character from pattern
        def pattern = antPattern.startsWith('/') ? antPattern : '/'+antPattern

        // deal with special regex-characters that should be interpreted as literals
        '\\.+[]^${}|()'.toCharArray().each{ c ->
            pattern = pattern.replace(''+c, '\\'+c)
        }
        pattern = pattern.replace('?', '.') // ? is a single-char wildcard

        // deal with ant-style wildcards
        StringBuffer result = new StringBuffer()
        result.append("^")
        def m = (pattern =~ '\\*\\*/|\\*\\*|\\*')
        while (m) {
            def token = m.group()
            def replacement;
            if (token == '**/') {
                replacement = '.*(?<=/)'
            }
            else if (token == '**') {
                replacement = '.*'
            }
            else {
                replacement = '[^/]*'
            }
            m.appendReplacement(result, Matcher.quoteReplacement(replacement))
        }
        m.appendTail(result)
        result.append("\$")
        def flags = 0
        if (isWindows) {
            flags |= Pattern.CASE_INSENSITIVE
        }
        return Pattern.compile(result.toString(), flags)
    }
}


class TestNGHandler extends DefaultHandler {
    final SUCCESS = "success"
    final FAILURE = "failure"
    final ERROR = "error"

    final resultMap = buildResultMap()

    def filename
    def onNewTest
    
    def inTest
    def inTestMethod
    def inMessage

    def testName
    def testResult
    def testDuration
    def testId

    def unitTestClassName
    def unitTestName

    def messages = []
    def currentText = new StringBuilder()

    TestNGHandler(def filename) {
        this.filename = filename
    }

    void startElement(String ns, String localName, String qName, Attributes attrs) {
        switch (qName) {
            case 'test':
                inTest = true
                unitTestName = attrs.getValue('name')
                break;

            case 'class':
                if (inTest) {
                    unitTestClassName = attrs.getValue('name')
                }
                break;

            case 'test-method':
                inTestMethod = true
                testId = attrs.getValue('signature')
                testName = attrs.getValue('name')
                testResult = toTestResult(attrs.getValue('status'))
                testDuration = attrs.getValue('duration-ms')
                break;

            case 'exception':
                if (inTestMethod) {
                    messages << 'Exception: ' + attrs.getValue('class')
                }
                break;

            case 'message':
                if (inTestMethod) {
                    inMessage = true
                }
                break;

            case 'full-stacktrace':
                if (inTestMethod) {
                    inMessage = true
                }
                break;
        }
    }

    void endElement(String ns, String localName, String qName) {
        switch (qName) {
            case 'test':
                inTest = false
                break;

            case 'test-method':
                onNewTest(testName, unitTestClassName, testResult, testDuration, messages)
                inTestMethod = false
                messages = []
                break;

            case 'message':
                if (inTestMethod) {
                    messages << 'Message: ' + currentText.toString().trim()
                    currentText = new StringBuilder()
                    inMessage = false;
                }
                break;

            case 'full-stacktrace':
                if (inTestMethod) {
                    messages << 'Full StackTrace: ' + currentText.toString().trim()
                    currentText = new StringBuilder()
                    inMessage = false;
                }
                break;
        }
    }

    void characters(char[] ch, int start, int length) {
        if (inMessage) {
            currentText.append(ch, start, length)
        }
    }

    def toTestResult(testOutcome) {
        return resultMap[testOutcome]
    }

    def buildResultMap() {
        def map = [
            'PASS': SUCCESS,
            'FAIL': FAILURE]
        return map.asImmutable()
    }

    def readTests(Closure closure) {
        onNewTest = closure
        def reader = SAXParserFactory.newInstance().newSAXParser().XMLReader
        def suiteFile = (filename instanceof File) ? filename : new File(filename)
        suiteFile.withInputStream { stream ->
            reader.setContentHandler(this)
            reader.parse(new InputSource(stream))
        }
    }
}

class EndParseEarlyException extends SAXException {}

class SuiteHandler extends DefaultHandler {

    def filename
    def suiteName

    SuiteHandler(filename) {
        this.filename = filename
    }

    void startElement(String ns, String localName, String qName, Attributes attrs) {
        if (qName == 'suite') {
            suiteName = attrs.getValue('name')
            throw new EndParseEarlyException()
        }
    }

    def getSuiteName() {
        def reader = SAXParserFactory.newInstance().newSAXParser().XMLReader
        try {
            def suiteFile = (filename instanceof File) ? filename : new File(filename)
            suiteFile.withInputStream { stream ->
                reader.setContentHandler(this)
                reader.parse(new InputSource(stream))
            }
        }
        catch (EndParseEarlyException swallow) {
        }
        return suiteName
    }
}

def ahpTool = new AHPTool();
def props = ahpTool.getStepProperties();

final def workDir = new File('.').canonicalFile
def baseDir = props['baseDir']
def reportName = props['reportName']
def includes = props['includes']
def excludes = props['excludes']

println "Base Directory: $baseDir"
println "Report Name: $reportName"
println "Includes: $includes"
println "Excludes: $excludes"

def fileSet = new FileSet(baseDir)
fileSet.include(includes)
fileSet.exclude(excludes)
def reportFiles = fileSet.files()

def builder = new StreamingMarkupBuilder()
def xml = builder.bind() {
    'test-report'(name: reportName, type: 'TestNG', 'job-id': System.getenv('AH_JOB_ID')) {
        reportFiles.each { file ->
            println("Parsing " + file.absolutePath)
            def suiteName = new SuiteHandler(file).getSuiteName()
            'test-suite'(name: suiteName) {
                def handler = new TestNGHandler(file)
                handler.readTests { testName, testClass, testResult, testDuration, messages ->
                    test(name: testName, 'class-name': testClass, result: testResult, time: testDuration) {
                        messages.each {
                            message(it)
                        }
                    }
                }
            }
        }
    }
}

if (reportFiles.size() > 0) {
    new CommandHelper(workDir).runCommand("Uploading to AHP", ['ahptool', 'addTests', '-report', reportName, '-']) { process ->
        process.withWriter { it << xml } 
        process.waitForProcessOutput(System.out, System.out)
    }
}
else {
    println 'Did not find any reports to publish!'
}
